/*
 * Copyright 2016-2017 gtalent2@gmail.com
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "addresses.hpp"
#include "panic.hpp"

namespace nostalgia {
namespace core {

struct HeapSegment {
	size_t size;
	uint8_t inUse;
	HeapSegment *next;
};

static HeapSegment *_heapIdx = nullptr;

void initHeap() {
	_heapIdx = (HeapSegment*) MEM_WRAM_END;
	// set size to half of WRAM
	_heapIdx->size = (MEM_WRAM_END - MEM_WRAM_BEGIN) / 2;
	_heapIdx->next = nullptr;
	_heapIdx->inUse = false;
}

}
}

using namespace nostalgia::core;

void *operator new(size_t sz) {
	// add space for heap segment header data
	sz += sizeof(HeapSegment);
	auto out = _heapIdx;
	while (out && out->size < sz) {
		out = out->next;
	}

	// panic if the allocation failed
	if (out == nullptr) {
		panic("Heap allocation failed");
	}

	// update size for the heap segment now that it is to be considered
	// allocated
	out->size = sz;
	out->next = (HeapSegment*) (((uint8_t*) out) + sz);
	out->inUse = true;

	auto hs = *_heapIdx;
	hs.size -= sz;
	if (hs.size == 0) {
		_heapIdx = hs.next;
	} else {
		_heapIdx = (HeapSegment*) (((uint8_t*) _heapIdx) - sz);
		*_heapIdx = hs;
	}

	return out;
}

void operator delete(void *ptr) {
	HeapSegment *prev = nullptr;
	HeapSegment *current = _heapIdx;
	while (current && current != ptr) {
		prev = current;
		current = current->next;
	}

	// ptr was found as a valid memory allocation, deallocate it
	if (current) {
		// mark as not in use
		prev->inUse = false;

		// join with next if next is also unused
		if (current->next && !current->next->inUse) {
			current->size += current->next->size;
			current->next = current->next->next;
		}

		// join with prev if prev is also unused
		if (prev && !prev->inUse) {
			prev->size += current->size;
			prev->next = current->next;
		}
	}
}

